;; Apple-1 6800 Pi
;;
;; Egan Ford (egan@sense.net)
;;
;; Sept, 10 2012
;;
;; Apple-1 6800 Pi can compute 1000 decimal digits of Pi.
;;
;; All Apple-1 Pi computation is base 256 and then converted to base 10 for
;; display (technically base 100, but they both display the same).  Arrays
;; of bytes are used to represent multiprecision numbers.  In the comments
;; below "array" refers to a big endian multiprecision number.
;;

;; TODO: optimize div8
;; consider using begloop instead of xreg for readability
;; update/verify all comments
;; headers for all sub and macros



	cpu	6800		; That's right, a 6800-based Apple-1

;; Apple-1 variables/vectors

echo	=	$FFB4		; character out sub
prbyte	=	$FFA2		; print a hex byte
warm	=	$FF21		; warm boot


;; my global variables/vectors

ptr_a	=	$0		; $0-$1 16 bit generic pointer
n	=	$2		; $2-$3 16 bit counter
carry_mp =	$4		; $4-$4  8 bit multiprecision carry
srcptr	=	$4		; $4-$5 16 bit stack pointers used for multi-
dstptr	=	srcptr		;              array operations (add_mp,
				;              sub_mp, copy_mp)
				;              32/16 div
dividend =	$4              ; $4-$5 16 bit dividend (hi) unused
				; $6-$7 16-bit dividend (lo)
divisor	=	$8		; $8-$9 16 bit
temp	=	$7		; $7-$7  8 bit for multi100 code
x2	=	$A		; $A-$A  8 bit cached atan square
regb	=	$B		; $B-$B  8 bit backup of regb
stkptr	=	$C		; $C-$D 16-bit stack pointer backup
xreg	=	$E		; $E-$F 16-bit x reg pointer
begloop	=	xreg
endloop	=	$10		; $10-$11 16-bit x reg pointer
dlength	=	$12		; $12-$13 16-bit number of digits (R/W)
squares	=	$14		; $14-$23 16-byte table of squares
monitor	=	$24		; $24-$2B reserved for monitor


;; my constants (R/O)

dec_len	=	1000		; 1000 decimal digits
bin_len	=	418		; ceil(1001 / log(256)) + 1 (+ 1 if odd) = 418


;; start of global macros

mset	macro	ptr, value	; a = x
	ldx	ptr
	ldab	value
	jsr	set_mp
	endm

mprint	macro	ptr
	ldx	ptr
	jsr	print_mp
	endm

mdiv	macro	ptr, value	; a = a / x
	ldx	ptr
	ldab	value
	jsr	div_mp
	endm

mdiv16	macro	ptr, value	; a = a / x
	ldx	value
	stx	divisor
	ldx	ptr
	jsr	div16_mp
	endm

mcopy	macro	dst, src	; a = b
	ldx	src
	stx	srcptr
	ldx	dst
	jsr	copy_mp
	endm

madd	macro	dst, src	; a = a + b
	ldx	src
	stx	srcptr
	ldx	dst
	jsr	add_mp
	endm

msub	macro	dst, src	; a = a - b
	ldx	dst
	stx	dstptr
	ldx	src
	jsr	sub_mp
	endm

masl	macro	ptr		; a = a * 2
	ldx	ptr
	jsr	asl_mp
	endm

matan	macro	ptr, value	; a = atan(1/x)
	ldx	ptr
	ldab	value
	jsr	atan_mp
	endm

add16	macro	a, b		; a = a + b, 4+4+5+4+4+5 = 26 cycles for extnd
	ldaa	b+1		; lo bits
	adda	a+1
	staa	a+1
	ldaa	b		; hi bits
	adca	a		; with carry
	staa	a
	endm

;; end of macros


;; start of real code

	.org	$280		; start here

	ldx	#pitext		; print "1000 DIGITS ..."
	jsr	print
	jsr	pi		; compute pi
	mprint	mp_a		; print it
	jsr	crout		; print CR
	jmp	warm		; all done, jmp to monitor


;; end of main code, start of non-zp allocations

arraylen:
	.byt	bin_len >> 8, bin_len & $FF	; length of base 256 array
arrayend:					; length of base 256 array
	.byt	(bin_len-1) >> 8
	.byt	(bin_len-1) & $FF
declen:	.byt	dec_len >> 8, dec_len & $FF	; decimal length of digits
						; mp pointers
mp_a:	.byt	mp_a_array >> 8, mp_a_array & $FF
mp_b:	.byt	mp_b_array >> 8, mp_b_array & $FF
mp_x:	.byt	mp_x_array >> 8, mp_x_array & $FF
mp_y:	.byt	mp_y_array >> 8, mp_y_array & $FF

sqrtbl:	.byt	0, 1, 4, 9, 16, 25, 36, 49, 64, 81, 100
	.byt	121, 144, 169, 196, 225

pitext:	.byt	"\r\r1000 DIGITS OF PI = \0"


;; start of subs

load_squares:
	sts	stkptr		; save stack pointer
	lds	#sqrtbl		; load stack pointer with sqr table
	des			; dec stack pointer (just the way it works)
	ldx	#0		; acctually going to use x for offset
	ldab	#16		; number of squares (0 - 15)
/	pula			; get square
	staa	squares,x	; store in direct page for speed and
	inx			;   well that is the only way it works
	decb			; all done?
	bne	-
	lds	stkptr		; restore stack pointer
	rts

pi:
	jsr	load_squares	; build square table
	matan	mp_a,#5		; a = atan(1/5)
	masl	mp_a		; a = a * 4
	masl	mp_a
	matan	mp_b,#239	; b = atan(1/239)
	msub	mp_a,mp_b	; a = a - b
	masl	mp_a		; a = a * 4
	masl	mp_a
        rts


atan_mp:
	stx	ptr_a		; store x in ptr_a
	stab	regb		; store b in regb

	ldaa	#1		; n = 1 big endian
	staa	n+1
	ldaa	#0
	staa	n+0
	staa	x2		; clear x2, hack used for getting x2
				; note regb MUST follow x2 in memory

	mset	mp_x,#1		; x = 1
	mdiv	mp_x,regb	; x = x / b
	mcopy	ptr_a,mp_x	; a = x
				; square x if x < 16
	ldaa	regb
	cmpa	#16
	bcc	+		; > 16
	ldx	x2		; 
	ldaa	squares,x
/	staa	x2

atan_mp_loop:			; main loop
	inc	n+1		; n now even
	bne	+		; n not zero
	inc	n+0		; n rolled to 0, inc n (hi)
/	inc	n+1		; n to be odd, save to inc (cannot roll from
				;   even to odd)
	ldaa	regb		; x = x / regx ^ 2
	cmpa	#16
	bcs	+		; x already x*x, one div required
	mdiv	mp_x,x2		; x >= 16, then do two div
/	mdiv	mp_x,x2		;   (faster than 16 bit div)

	mcopy	mp_y,mp_x	; y = x

	ldaa	n+0		; if n < 256 then div else div16
	bne	+		; >= 256 use 16 bit div
	mdiv	mp_y,n+1	; < 256 use 8 bit div
	bcs	atan_mp_end	; dividend = 0, done
	bcc	++		; still working on it
/	mdiv16	mp_y,n		; div16
	bcs	atan_mp_end	; dividend = 0, done

/	ldaa	n+1		; add or sub
	anda	#2
	beq	+		; add it
	msub	ptr_a,mp_y	; a = a - y 
	jmp	atan_mp_loop
/	madd	ptr_a,mp_y	; a = a + y 
	jmp	atan_mp_loop
atan_mp_end:
	rts


copy_mp:			; a = b (dest = src)
	sts	stkptr		; backup stack pointer
	lds	srcptr		; use stack for source
	des			; dec stack back one (just the way sp works)
	stx	endloop		; store x in xreg
	add16	endloop,arrayend; endloop = endloop + arrayend
	dex			; start back one
	bra	++		; Duff device hardcoded for fixed arraylen
/	inx			; next x
	pula			; read b
	staa	,x		; write to a
	inx			; next x
	pula			; read b
	staa	,x		; write to a
/	inx			; next x
	pula			; read b
	staa	,x		; write to a
	inx			; next x
	pula			; read b
	staa	,x		; write to a
	cpx	endloop		; all done?
	bne	--
	lds	stkptr		; restore stack pointer
	rts


add_mp:
	sts	stkptr		; backup stack pointer
	lds	srcptr		; use stack for source
	des			; dec stack back one (just the way sp works)
	stx	endloop		; store x in xreg
	add16	endloop,arrayend; endloop = endloop + arrayend
	clc			; clear carry
	dex			; start back one
	bra	++		; Duff device hardcoded for fixed arraylen
/	inx			; next x
	pula			; read b
	adca	,x		; a = a + b
	staa	,x		; write to a
	inx			; next x
	pula			; read b
	adca	,x		; a = a + b
	staa	,x		; write to a
/	inx			; next x
	pula			; read b
	adca	,x		; a = a + b
	staa	,x		; write to a
	inx			; next x
	pula			; read b
	adca	,x		; a = a + b
	staa	,x		; write to a
	cpx	endloop		; all done?
	bne	--
	lds	stkptr		; restore stack pointer
	rts


sub_mp:
	sts	stkptr		; backup stack pointer
	lds	dstptr		; use stack for dst
	des			; dec stack back one (just the way sp works)
	stx	endloop		; store x in xreg
	add16	endloop,arrayend; endloop = endloop + arrayend
	clc			; clear carry
	dex			; start back one
	bra	++		; Duff device hardcoded for fixed arraylen
/	inx			; next x
	pula			; read a
	sbca	,x		; a = a - b
	psha			; store a
	ins			; inc stack for next a
	inx			; next x
	pula			; read a
	sbca	,x		; a = a - b
	psha			; store a
	ins			; inc stack for next a
/	inx			; next x
	pula			; read a
	sbca	,x		; a = a - b
	psha			; store a
	ins			; inc stack for next a
	inx			; next x
	pula			; read a
	sbca	,x		; a = a - b
	psha			; store a
	ins			; inc stack for next a
	cpx	endloop		; all done?
	bne	--
	lds	stkptr		; restore stack pointer
	rts
	rts

asl_mp:
	stx	endloop		; store x in xreg
	add16	endloop,arrayend; endloop = endloop + arrayend
	clc			; clear carry
	dex			; start back one
	bra	++		; Duff device hardcoded for fixed arraylen
/	inx			; next x
	rol	,x		; a = a * 2
	inx			; next x
	rol	,x		; a = a * 2
/	inx			; next x
	rol	,x		; a = a * 2
	inx			; next x
	rol	,x		; a = a * 2
	cpx	endloop		; all done?
	bne	--
	rts


skipzeros:
	stx	endloop		; 6
	stx	xreg		; 6
	add16	xreg,arraylen	; 26
				; total 26 + 6 + 6 + 5 = 43
	ldx	xreg		; x set to end of array
	bra	++		; Duff device hardcoded for fixed arraylen
/	dex
	ldaa	,x
	bne	even
	dex
	ldaa	,x
	bne	odd
/	dex
	ldaa	,x
	bne	even
	dex
	ldaa	,x
	bne	odd
	cpx	endloop
	bne	--
	sec			; all zeros
	rts
odd:	inx			; make even
even:	inx			; offset +1 because we lead with dex
	clc
	rts

div8	macro			; need to optimize this
	asla
	rept	8		; unroll 8 times
	rolb
	bcs	m0
	cmpb	divisor
	bcc	m0
	clc
	bcc	m1
m0:	subb	divisor
	sec
m1:	rola
	endm
	endm

div_mp:
	stab	divisor
	jsr	skipzeros
	bcc	+
	rts
/	clrb			; set remainder to 0
/	dex
	ldaa	,x
	div8
	staa	,x
	dex
	ldaa	,x
	div8
	staa	,x
	cpx	endloop
	beq	+		; long branch
	jmp	-
/	clc
	rts


div16	macro			; array little endian, 
				;   needed to swap dividend+2 and +3
	rept	16		; unroll 16 times
	asl	dividend+2	; 6
	rol	dividend+3	; 6
	rolb			; 2
	rola			; 2
	subb	divisor+1	; 3
	sbca	divisor+0	; 3
	bcs	m0		; 4
	inc	dividend+2	; 6 
	jmp	m1		; 3 = 35 cycles with sub
m0:	addb	divisor+1	; 3
	adca	divisor+0	; 3 = 32 cycles no sub
m1:
	endm
	endm

div16_mp:
	jsr	skipzeros
	bcc	+
	rts
/	sts	stkptr		; save stack pointer
	clra			; set remainder to 0
	clrb
/	dex
	dex	
	lds	,x
	sts	dividend+2
	div16
	lds	dividend+2
	sts	,x
	cpx	endloop
	beq	+		; long branch
	jmp	-
/	clc
	lds	stkptr		; restore stack pointer
	rts


print_mp:
				; setup start/end for x100 loop
	stx	xreg		; backup start of array pointer (LSB)
	stx	endloop		; store x in xreg
	add16	endloop,arrayend; endloop = endloop + arrayend

				; div declen by 2 since we are printing base100
	ldx	declen		; backup declen to dlength for multiple prints
	stx	dlength
	lsr	dlength		; lsr hi
	ror	dlength+1	; ror lo

				; print left of decimal point
	ldx	endloop		; get last digit (MSB) in array (little endian)
	ldaa	,x
	jsr	bintobcd	; convert from hex to dec
	cmpa	#9		; if <= 9 skip leading zero
	bls	+		;   branch to print single digit ascii
	jsr	prbyte		; otherwise just print it
	bra	++		; skip to print period
/	clc			; no leading zero code
	adca	#$B0		;   convert single digit to ascii
	jsr	echo		;   print it
/	ldaa	#'.'+$80	; print decimal point (fixed point math)
	jsr	echo		;   print it
	clr	,x		; zero out MSB

print_mp1:			; loop through rest of decimal digits
	; multi array x 100
	; loop from LSB to MSB
	; 16-bit product = array[i] * 100 + carry_mp;
	; array[i] = product lo
	; carry_mp = product hi
	;
	; C Algorithm:
	;
	;    while (j >= 0) {
	;       result.L = (long) number[j] * 100 + carry;
	;       number[j] = result.w.lo;
	;       carry = result.w.hi;
	;       j--;
	;    }

mult100	macro			; 100 = 1100100, so a = ((x*2 + x)*8 + x) * 4
	ldab	#0		; MSB
	ldaa	,x		; LSB
	staa	temp		; save x
	asla			; x*2
	rolb
	adda	temp		; +x
	bcc	m1		; if carry
	incb			;   then inc MSB
m1:	asla			; *8
	rolb
	asla
	rolb
	asla
	rolb
	adda	temp		; +x
	bcc	m2
	incb
m2:	asla			; *4
	rolb
	asla
	rolb
	adda	carry_mp	; + carry from chain
	bcc	m3
	incb
m3:	staa	,x		; save it
	stab	carry_mp
	endm
				; mp array * 100
	clr	carry_mp	; zero out carry
	ldx	xreg		; forward loop
	dex			; start back one
	jmp	++		; Duff device hardcoded for fixed arraylen
/	inx			; next x
	mult100	
	inx			; next x
	mult100	
/	inx			; next x
	mult100	
	inx			; next x
	mult100	
	cpx	endloop		; all done?
	beq	+		; long branch (unroll too long)
	jmp	--
/
				; print MSB
	ldaa	,x		; get last digit (MSB) in array (little endian)
	jsr	bintobcd	; convert from hex to dec
	jsr	prbyte		; print it
	clr	,x		; zero out MSB

	ldx	dlength		; dlength = dlength - 1
	dex
	beq	+
	stx	dlength
	jmp	print_mp1
/	rts


; Description: BIN/HEX to BCD.  BCD = BIN + 6*int(BIN/10).
;
; Input:       A (BIN)
;
; Output:      A (BCD)
;
; Clobbered:   A, B

bintobcd:
	psha			; backup A
	clc			; clear carry (6800 is NOT the 6502)
	ldab	#-1		; start counter at -1
/	incb			; inc b (first loop = 0)
	sbca	#10		; a = a - 10
	bpl	-		; > 0?, then keep - 10s
	clc			; clear carry, the sbca < 0 should have set it
	pula			; restore A
	bra	++		; jump to decb
/	adca	#6		; a = a + 6
/	decb			; if a < 10, then b will be -1 and end loop
	bpl	--		;   otherwise add 6 for every 10
	rts


set_mp:
	stx	endloop		; store x in xreg
	add16	endloop,arrayend; endloop = endloop + arrayend
	dex			; start back one
	ldaa	#0
	bra	++		; Duff device hardcoded for fixed arraylen
/	inx			; next x
	staa	,x		; faster than clra
	inx			; next x
	staa	,x		; faster than clra
/	inx			; next x
	staa	,x		; faster than clra
	inx			; next x
	staa	,x		; faster than clra
	cpx	endloop		; all done?
	bne	--
	stab	,x		; write digit left of decimal
	rts


print:	ldaa	,x
	beq	+
	jsr	echo
	inx
	bra	print
/	rts


crout:	ldaa	#$8D
        jmp	echo


; data

mp_a_array:
	.org	*+bin_len
mp_b_array:
	.org	*+bin_len
mp_x_array:
	.org	*+bin_len
mp_y_array:
	.org	*+bin_len

end:

